@@ -38,6 +38,10 @@ static char sccsid[] = "@(#)cmds.c	5.26 (Berkeley) 3/5/91";
 /*
  * FTP User Program -- Command Routines.
  */
+#ifdef _WIN32
+#include <win-mac.h>
+#endif
+
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
@@ -66,6 +70,8 @@ static char sccsid[] = "@(#)cmds.c	5.26 (Berkeley) 3/5/91";
 #include <ctype.h>
 #include <time.h>
 
+#include <k5-platform.h>
+
 #ifdef HAVE_GETCWD
 #define getwd(x) getcwd(x,MAXPATHLEN)
 #endif
@@ -182,7 +188,7 @@ void setpeer(argc, argv)
 		form = FORM_N;
 		mode = MODE_S;
 		stru = STRU_F;
-		(void) strcpy(bytename, "8"), bytesize = 8;
+		(void) strlcpy(bytename, "8", sizeof(bytename)), bytesize = 8;
 		if (autoauth) {
 			if (do_auth() && autoencrypt) {
  				clevel = PROT_P;
@@ -1168,7 +1174,7 @@ void status(argc, argv)
 	if (!proxy) {
 		pswitch(1);
 		if (connected) putchar('\n');
-		cstatus(argc,argv);
+		cstatus();
 		if (connected) putchar('\n');
 		pswitch(0);
 	}
@@ -1615,9 +1621,7 @@ void shell(argc, argv)
 		namep = strrchr(shellprog,'/');
 		if (namep == NULL)
 			namep = shellprog;
-		(void) strcpy(shellnam,"-");
-		(void) strncat(shellnam, ++namep, sizeof(shellnam) - 1 - strlen(shellnam));
-		shellnam[sizeof(shellnam) - 1] = '\0';
+		(void) snprintf(shellnam, sizeof(shellnam), "-%s", ++namep);
 		if (strcmp(namep, "sh") != 0)
 			shellnam[0] = '+';
 		if (debug) {
