@@ -75,7 +75,7 @@ SOFTWARE.
 *         are noted in the comments section of each of the
 *         functions.
 *
-* rcsid: $Id: custom.cpp 17533 2005-12-02 10:29:29Z jaltman $
+* rcsid: $Id: custom.cpp 20119 2007-10-12 23:08:30Z jaltman $
 **************************************************************/
 
 #pragma unmanaged
@@ -87,6 +87,9 @@ SOFTWARE.
 // linker stuff
 #pragma comment(lib, "msi")
 #pragma comment(lib, "advapi32")
+#if defined(_M_IA64) || defined(_M_AMD64)
+#pragma comment(lib, "bufferoverflowu")
+#endif
 
 
 void ShowMsiError( MSIHANDLE hInstall, DWORD errcode, DWORD param ){
@@ -548,7 +551,7 @@ MSIDLLEXPORT UninstallNsisInstallation( MSIHANDLE hInstall )
 	sInfo.cb = sizeof(sInfo);
 	sInfo.lpReserved = NULL;
 	sInfo.lpDesktop = _T("");
-	sInfo.lpTitle = _T("Foo");
+	sInfo.lpTitle = _T("NSIS Uninstaller for Kerberos for Windows");
 	sInfo.dwX = 0;
 	sInfo.dwY = 0;
 	sInfo.dwXSize = 0;
@@ -575,11 +578,24 @@ MSIDLLEXPORT UninstallNsisInstallation( MSIHANDLE hInstall )
 		NULL,
 		&sInfo,
 		&pInfo)) {
-			pInfo.hProcess = NULL;
-			pInfo.hThread = NULL;
-			rv = 40;
-			goto _cleanup;
-		};
+            DWORD lastError = GetLastError();
+            MSIHANDLE hRecord;
+
+            hRecord = MsiCreateRecord(4);
+            MsiRecordClearData(hRecord);
+            MsiRecordSetInteger(hRecord, 1, ERR_NSS_FAILED_CP);
+            MsiRecordSetString(hRecord, 2, strPathUninst);
+            MsiRecordSetInteger(hRecord, 3, lastError);
+
+            MsiProcessMessage( hInstall, INSTALLMESSAGE_ERROR, hRecord );
+	
+            MsiCloseHandle( hRecord );
+
+            pInfo.hProcess = NULL;
+            pInfo.hThread = NULL;
+            rv = 40;
+            goto _cleanup;
+        };
 
 	// Create a job object to contain the NSIS uninstall process tree
 
@@ -627,8 +643,8 @@ _cleanup:
 	if(hJob) CloseHandle(hJob);
 	if(strPathUninst) delete strPathUninst;
 
-	if(rv != ERROR_SUCCESS) {
-        ShowMsiError( hInstall, ERR_NSS_FAILED, rv );
+	if(rv != ERROR_SUCCESS && rv != 40) {
+            ShowMsiError( hInstall, ERR_NSS_FAILED, rv );
 	}
 	return rv;
 }
@@ -704,10 +720,12 @@ DWORD InstNetProvider(MSIHANDLE hInstall, int bInst) {
     dwSize = 0;
     CHECK(rv = RegQueryValueEx( hkOrder, STR_VAL_ORDER, NULL, NULL, NULL, &dwSize ) );
 
-    strOrder = new TCHAR[ (dwSize + STR_SERVICE_LEN) * sizeof(TCHAR) ];
+    strOrder = new TCHAR[ (dwSize + STR_SERVICE_LEN + 4) * sizeof(TCHAR) ];
 
     CHECK(rv = RegQueryValueEx( hkOrder, STR_VAL_ORDER, NULL, NULL, (LPBYTE) strOrder, &dwSize));
 
+    strOrder[dwSize] = '\0';	/* reg strings are not always nul terminated */
+
     npi_CheckAndAddRemove( strOrder, STR_SERVICE , bInst);
 
     dwSize = (lstrlen( strOrder ) + 1) * sizeof(TCHAR);
